from collections import defaultdict
from typing import DefaultDict, List, Sequence, Tuple


class quickFilter:
  """
  custom fuzzy filter that uses 1,2,3-grams to filter and sort
  several strings by their similarity with a given pattern
  """
  _elements: Sequence[str]
  _ngrams:   DefaultDict[str, Sequence[Tuple[int,int]]]

  def __init__(self, elems: Sequence[str]):
    self._elements = elems
    low_elems = [s.lower() for s in elems]
    results = [
      (sub, i, score)
      for i, elem in enumerate(low_elems)
      for sub, score in self._get_grams(elem)
    ]
    best = defaultdict(lambda: defaultdict(lambda: 0))
    for sub, i, score in results:
      best[sub][i] = max(best[sub][i], score)
    ngrams = defaultdict(lambda: [])
    for (sub, i), score in best.items():
      ngrams[sub].append((i, score))
    self._ngrams = ngrams # type: ignore

  def _get_grams(self, pattern: str) -> Sequence[Tuple[str,int]]:
    '''
    Compute all 1,2,3-grams of pattern and their score, possibly repeating.
    The output size is bounded by 3*len(pattern)
    Additional score is given if the gram comes after space, dash or underscore.
    '''
    n = len(pattern)
    out = []
    for i in range(n):
      isStart = (i==0) or (pattern[i-1] in (' ', '-', '_'))
      for j in range(i+1, min(i+3, n)):
        sub = pattern[i:j]
        score = 1 + isStart
        out.append((sub, score))
    return out

  def getFiltered(self, pattern:str) -> List[int]:
    '''
    Get the indices of elements that match the pattern in fuzzy manner
    '''
    if len(pattern) == 0:
      out = list(range(len(self._elements)))
    else:
      pattern = pattern.lower()
      scored = defaultdict(lambda: 0)
      for sub, _ in self._get_grams(pattern):
        for i, score in self._ngrams[sub]:
          scored[i] += score
      ranking = sorted(sorted.items(), key=(lambda _,score: -score))) # type: ignore
      threshold = max(1, len(pattern)*3-4)
      out = [i for i, score in ranking if score>threshold]
    return out


'''

package main

import (
  "fmt"
  "io/ioutil"
  "log"
  "os"
  "path/filepath"
  "sort"
  "strings"

  "fyne.io/fyne/v2"
  "fyne.io/fyne/v2/app"
  "fyne.io/fyne/v2/container"
  "fyne.io/fyne/v2/driver/desktop"
  "fyne.io/fyne/v2/widget"
)

var ShortcutAltUp = desktop.CustomShortcut{KeyName: fyne.KeyUp, Modifier: desktop.AltModifier}
var ShortcutCtrlBackspace = desktop.CustomShortcut{KeyName: fyne.KeyBackspace, Modifier: desktop.ControlModifier}

type escapeEntry struct {
  widget.Entry
  onEnter        func()
  onArrowUp      func()
  onArrowDown    func()
  onAltArrowUp   func()
  onAltArrowDown func()
}

func (e *escapeEntry) onEsc() {
  fmt.Println(e.Entry.Text)
  e.Entry.SetText("")
}

func newEscEntry() *escapeEntry {
  entry := &escapeEntry{}
  entry.ExtendBaseWidget(entry)
  return entry
}

func (e *escapeEntry) TypedKey(key *fyne.KeyEvent) {
  switch key.Name {
  case fyne.KeyEscape:
    e.onEsc()
  case fyne.KeyReturn:
    e.onEnter()
  case fyne.KeyUp:
    e.onArrowUp()
  case fyne.KeyDown:
    e.onArrowDown()
  default:
    e.Entry.TypedKey(key)
  }
}
func (e *escapeEntry) TypedShortcut(s fyne.Shortcut) {
  if _, ok := s.(*desktop.CustomShortcut); !ok {
    e.Entry.TypedShortcut(s)
    return
  }
  name := s.ShortcutName()
  switch name {
  case ShortcutAltUp.ShortcutName():
    e.onAltArrowUp()
  case ShortcutCtrlBackspace.ShortcutName():
    e.Entry.SetText("")
  default:
    log.Println("Shortcut typed:", name)
  }
}

type myPath struct {
  info      os.FileInfo
  name      string
  path      string
  resolved  string
  isSymlink bool
  isDir     bool
  icon      string
}

func newMyPath(cwd, name string) *myPath {
  e := new(myPath)
  e.name = name
  e.path = fmt.Sprintf("%s/%s", cwd, name)
  info, _ := os.Lstat(e.path)
  e.info = info
  resolved, err := filepath.EvalSymlinks(e.path)
  e.resolved = resolved
  rinfo, err := os.Lstat(resolved)
  e.isDir = (err == nil) && rinfo.IsDir()
  e.isSymlink = (e.path != resolved)
  e.icon = "f"
  if e.isDir {
    e.icon = "D"
  }
  if e.isSymlink {
    e.icon = "s" + e.icon
  }
  e.icon = fmt.Sprintf("[%s]", e.icon)
  return e
}

type dirList struct {
  wPwd     *widget.Button
  wCwd     *widget.Label
  wFilter  *escapeEntry
  wVBox    *fyne.Container
  objects  []*widget.Label
  filter   *quickFilter
  root     string
  complete []string
  filtered []int
  shown    []int
  nshown   int
  offset   int
  fileInfo map[string]*myPath
  selected int
}

func NewDirList(root string) *dirList {
  e := new(dirList)
  e.nshown = 10
  e.wCwd = widget.NewLabel("")
  e.wPwd = widget.NewButton("Parent (alt+up)", e.enterParent)
  e.wFilter = newEscEntry()
  e.wFilter.Entry.SetPlaceHolder("Filter...")
  e.wFilter.Entry.OnChanged = e.applyFilter
  e.wFilter.onEnter = e.onEnter
  e.wFilter.onArrowUp = e.onArrowUp
  e.wFilter.onArrowDown = e.onArrowDown
  e.wFilter.onAltArrowUp = e.enterParent
  e.wVBox = container.NewVBox()
  e.enterDir(root)
  return e
}

func (e *dirList) enterDir(root string) {
  e.root = root
  e.wCwd.SetText(root)
  e.fileInfo = map[string]*myPath{}
  e.complete = []string{}
  files, err := ioutil.ReadDir(root)
  if err != nil {
    if root == "/" {
      log.Fatal(err)
    } else {
      e.enterDir("/")
    }
  }
  for _, file := range files {
    name := file.Name()
    e.fileInfo[name] = newMyPath(root, name)
    e.complete = append(e.complete, name)
  }
  sort.SliceStable(e.complete, func(i, j int) bool {
    return strings.ToLower(e.complete[i]) < strings.ToLower(e.complete[j])
  })
  e.filter = newQuickFilter(e.complete)
  e.wFilter.SetText("")
  e.applyFilter(e.wFilter.Text)
}

func (e *dirList) applyFilter(filter string) {
  showHidden := (len(filter) > 0 && filter[0] == '.')
  e.filtered = []int{}
  for _, i := range e.filter.getFiltered(filter) {
    name := e.complete[i]
    isHidden := (name[0] == '.')
    if !isHidden || showHidden {
      e.filtered = append(e.filtered, i)
    }
  }
  e.offset = 0
  e.selected = -1
  if len(filter) > 0 && len(e.filtered) > 0 {
    e.selected = 0
  }
  e.computeShown()
}

func (e *dirList) computeShown() {
  e.shown = []int{}
  for i := e.offset; i < len(e.filtered) && i < e.offset+e.nshown; i++ {
    e.shown = append(e.shown, e.filtered[i])
  }
  out := []fyne.CanvasObject{}
  same_out := []*widget.Label{}
  for _, ifile := range e.shown {
    name := e.complete[ifile]
    info := e.fileInfo[name]
    w := widget.NewLabel(fmt.Sprintf("%s %s", info.icon, info.name))
    out = append(out, w)
    same_out = append(same_out, w)
  }
  e.wVBox.Objects = out
  e.objects = same_out
  e.highlightSelected()
}

func (e *dirList) highlightSelected() {
  for i, obj := range e.objects {
    obj.TextStyle.Bold = (e.offset+i == e.selected)
  }
  e.wVBox.Refresh()
}

func (e *dirList) onEnter() {
  if e.selected == -1 {
    return
  }
  name := e.complete[e.filtered[e.selected]]
  info := e.fileInfo[name]
  if info.isDir {
    e.enterDir(info.path)
  } else {
    log.Print(info.path)
  }
}

func (e *dirList) enterParent() {
  root := strings.TrimRight(e.root, "/")
  seps := strings.Split(root, "/")
  seps = seps[:len(seps)-1]
  parent := strings.Join(seps, "/")
  e.enterDir(parent)
}

func (e *dirList) onArrowUp() {
  if e.selected <= -1 {
    return
  }
  e.selected = e.selected - 1
  if e.selected < e.offset && e.selected >= 0 {
    e.offset = e.offset - 1
    e.computeShown()
  }
  e.highlightSelected()
}

func (e *dirList) onArrowDown() {
  if e.selected+1 >= len(e.filtered) {
    return
  }
  e.selected = e.selected + 1
  if e.selected >= e.offset+e.nshown {
    e.offset = e.offset + 1
    e.computeShown()
  }
  e.highlightSelected()
}

func main() {
  a := app.New()
  w := a.NewWindow("Hello")

  //filter := newEscEntry()
  //filter.Entry.SetPlaceHolder("Filter...")
  //command := newEscEntry()
  //filter.Entry.SetPlaceHolder("Command...")

  root := "/home/carlos"
  dirs := NewDirList(root)

  ctrlTab := desktop.CustomShortcut{KeyName: fyne.KeyTab, Modifier: desktop.ControlModifier}
  w.Canvas().AddShortcut(&ctrlTab, func(shortcut fyne.Shortcut) {
    fmt.Println("We tapped Ctrl+Tab")
  })

  //hello := widget.NewLabel("Hello Fyne!")
  w.SetContent(container.NewVBox(
    dirs.wPwd,
    dirs.wCwd,
    dirs.wFilter,
    dirs.wVBox,
    /*widget.NewButton("Hi!", func() {
      hello.SetText("Welcome :)")
      //dirs.Objects = append(dirs.Objects, widget.NewLabel("/home"))
    }),
    */
    //hello,
  ))
  w.Canvas().Focus(dirs.wFilter)
  w.ShowAndRun()
}
'''


import heapq
from typing import Any, Dict, List, Tuple, Union
'''
Some functions are copied from heapq
because they start with underscore and may
be eliminated in the future
'''


def _heappop_max(heap):
    """Maxheap version of a heappop."""
    lastelt = heap.pop()  # raises appropriate IndexError if heap is empty
    if heap:
        returnitem = heap[0]
        heap[0] = lastelt
        _siftup_max(heap, 0)
        return returnitem
    return lastelt


def _siftdown_max(heap, startpos, pos):
    'Maxheap variant of _siftdown'
    newitem = heap[pos]
    # Follow the path to the root, moving parents down until finding a place
    # newitem fits.
    while pos > startpos:
        parentpos = (pos - 1) >> 1
        parent = heap[parentpos]
        if parent < newitem:
            heap[pos] = parent
            pos = parentpos
            continue
        break
    heap[pos] = newitem


def _heapify_max(x):
    """Transform list into a maxheap, in-place, in O(len(x)) time."""
    n = len(x)
    for i in reversed(range(n // 2)):
        _siftup_max(x, i)


def _siftup_max(heap, pos):
    'Maxheap variant of _siftup'
    endpos = len(heap)
    startpos = pos
    newitem = heap[pos]
    # Bubble up the larger child until hitting a leaf.
    childpos = 2 * pos + 1  # leftmost child position
    while childpos < endpos:
        # Set childpos to index of larger child.
        rightpos = childpos + 1
        if rightpos < endpos and not heap[rightpos] < heap[childpos]:
            childpos = rightpos
        # Move the larger child up.
        heap[pos] = heap[childpos]
        pos = childpos
        childpos = 2 * pos + 1
    # The leaf at pos is empty now.  Put newitem there, and bubble it up
    # to its final resting place (by sifting its parents down).
    heap[pos] = newitem
    _siftdown_max(heap, startpos, pos)


Comparable = Any


class MinHeap(List[Comparable]):

    def _heappop(self):
        return heapq.heappop(self)

    def _heappush(self, elem: Comparable):
        heapq.heappush(self, elem)

    def _heapify(self):
        heapq.heapify(self)


class MaxHeap(List[Comparable]):

    def _heappop(self):
        return _heappop_max(self)

    def _heappush(self, elem: Comparable):
        self.append(elem)
        _siftdown_max(self, 0, len(self) - 1)

    def _heapify(self):
        _heapify_max(self)


Heap = Union[MinHeap, MaxHeap]


class _HeapDict(Dict[Any, Comparable]):
    '''
    Parent class for MinDict and MaxDict.
    Supports O(log n) lookup or elimination of heap root.
    '''
    heapsize_ignore_until = 200
    heapsize_max_factor = 4
    heap_class = MinHeap

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._heap: Heap = self.heap_class()
        self._init_heap()

    def _init_heap(self):
        elems = [(v, k) for k, v in self.items()]
        self._heap[:] = elems
        self._heap._heapify()

    def __setitem__(self, key: Any, value: Comparable):
        curr = self.get(key)
        found = (curr is not None) or (key in self)
        if not found or curr != value:
            super().__setitem__(key, value)
        if found and curr != value:
            self._heap._heappush((value, key))
            self._clean_heap()

    def _get_root(self):
        self._clean_heap()
        value, key = self._heap[0]
        assert self[key] == value
        return (key, value)

    def _pop_root(self):
        self._clean_heap()
        value, key = self._heap._heappop()
        assert self[key] == value
        del self[key]
        return (key, value)

    def _clean_heap(self):
        while len(self._heap):
            value, key = self._heap[0]
            curr = self.get(key)
            found = (curr is not None) or (key in self)
            if found and curr == value:
                break
            self._heap._heappop()

        too_large = max(
            self.heapsize_max_factor * len(self),
            self.heapsize_ignore_until,
        )
        if len(self._heap) > too_large:
            self._init_heap()


class MinDict(_HeapDict):
    '''
    Dictionary supporting O(log n) lookup or elimination
    of key with minimum value.
    '''
    heapsize_ignore_until = 200
    heapsize_max_factor = 4
    heap_class = MinHeap

    def get_min(self):
        return self._get_root()

    def pop_min(self):
        return self._pop_root()


class MaxDict(_HeapDict):
    '''
    Dictionary supporting O(log n) lookup or elimination
    of key with maximum value.
    '''
    heapsize_ignore_until = 200
    heapsize_max_factor = 4
    heap_class = MaxHeap

    def get_max(self):
        return self._get_root()

    def pop_max(self):
        return self._pop_root()



import functools
import weakref
from typing import Any, Callable, Optional, Type, TypeVar, Union, cast
from typing_extensions import Protocol
'''
Typing this library correctly is impossible right now.

The main issue is that we can not create types for callable objects with a partial signature, like "Callable[[SELF, ...], OUT]" for a normal method.

https://github.com/python/mypy/issues/5876

'''

_T = TypeVar('_T')


class _NormalMethod(Protocol[_T]):
    __name__: str
    __call__: Callable[..., _T]


_ClassMethod = _NormalMethod
_StaticMethod = _NormalMethod


class _Property(Protocol[_T]):
    __get__: Callable[..., _T]


class cached_property(property, _Property[_T]):
    '''
    decorator for converting a method into a cached property
    See https://stackoverflow.com/a/4037979/3671939
    This uses a modification:
    1. inherit from property, which disables setattr(instance, name, value)
        as it raises AttributeError: Can't set attribute
    2. use instance.__dict__[name] = value to fix
    '''

    def __init__(self, method: Callable[..., _T]):
        self._method = method

    def __get__(self, instance, _) -> _T:
        name = self._method.__name__
        value = self._method(instance)
        instance.__dict__[name] = value
        return value


def set_method(cls):
    '''decorator for adding or replacing a method of a given class'''
    return _set_method(cls, None)


def _set_method(cls, proxy: Callable = None):
    '''decorator for adding or replacing a method of a given class'''

    def decorator(method: _NormalMethod):
        assert hasattr(method, '__call__'), f'Not callable method: {method}'
        name: str = method.__name__
        method = proxy(method) if proxy else method
        setattr(cls, name, method)

    return decorator


def cached_method(maxsize=128, typed=False):
    '''decorator for converting a method into an lru cached method'''

    # https://stackoverflow.com/a/33672499/3671939
    def decorator(func: _NormalMethod[_T]) -> _NormalMethod[_T]:

        @functools.wraps(func)
        def wrapped_func(self, *args, **kwargs):
            # We're storing the wrapped method inside the instance. If we had
            # a strong reference to self the instance would never die.
            weak_self = weakref.ref(self)

            @functools.wraps(func)
            @functools.lru_cache(maxsize=maxsize, typed=typed)
            def _cached_method(*args, **kwargs):
                self = weak_self()
                assert self
                return func(self, *args, **kwargs)

            setattr(self, func.__name__, _cached_method)
            return _cached_method(*args, **kwargs)

        return wrapped_func

    return decorator


def set_cached_method(cls, maxsize=128, typed=False):
    '''decorator for adding or replacing a cached_method of a given class'''
    return _set_method(cls, cached_method(maxsize, typed))


def set_classmethod(cls):
    '''decorator for adding or replacing a classmethod of a given class'''
    return _set_method(cls, classmethod)


def set_staticmethod(cls):
    '''decorator for adding or replacing a staticmethod of a given class'''
    return _set_method(cls, staticmethod)


def set_property(cls):
    '''decorator for adding or replacing a property of a given class'''
    return _set_method(cls, property)


def set_cached_property(cls):
    '''decorator for adding or replacing a cached_property of a given class'''
    return _set_method(cls, cached_property)


set_cachedproperty = set_cached_property  # Previous versions

if False:

    class Test:
        zero = 0

        @cached_property
        def prop(self):
            print('x')
            return 1 + 1

    @set_method(Test)
    def f(self: Test):
        print(self.zero)

    t = Test()
    x = t.prop


from .methodtools import cached_property, cached_method
import warnings


class MissingPackage(UserWarning):
    pass


try:
    import ffmpeg
except ImportError:
    warnings.warn('pip3 install ffmpeg-python', MissingPackage)
try:
    import numpy as np
except ImportError:
    warnings.warn('pip3 install numpy', MissingPackage)
try:
    from scipy.io import wavfile
except ImportError:
    warnings.warn('pip3 install scipy', MissingPackage)

import os
import ffmpeg
import numpy as np
from scipy.io import wavfile


def ffmpeg_run(out):
    try:
        ffmpeg.run(out, capture_stdout=True, capture_stderr=True,
                   overwrite_output=True)
    except ffmpeg.Error as e:
        print(e.stderr.decode())
        raise
    return


class Sequence(np.ndarray):

    rate: float

    def __new__(cls, arr, rate: float):
        obj = np.asarray(arr).view(cls)
        obj.rate = rate  # (sample rate Hz)
        return obj

    @classmethod
    def from_wav(cls, file):
        rate, data = wavfile.read(file)
        return cls(data, rate=rate)

    def to_wav(self, file):
        return wavfile.write(file, rate=self.rate, data=self)

    @cached_property
    def duration(self):
        return len(self) / self.rate

    def _float_index(self, idx_float):
        idx_floor = np.floor(idx_float).astype(int)
        p_next = idx_float - idx_floor
        idx_next = np.minimum(idx_floor + 1, len(self) - 1)
        arr = self[idx_floor] * (1 - p_next) + self[idx_next] * p_next
        return self.rated(arr.astype(self.dtype))

    def rated(self, arr):
        return self.__class__(arr, self.rate)

    def clip_cast(self, arr):
        if arr.dtype != self.dtype:
            lo = np.ma.maximum_fill_value(self.dtype)
            hi = np.ma.minimum_fill_value(self.dtype)
            arr = np.clip(arr, lo, hi).astype(self.dtype)
        return self.__class__(arr, self.rate)

    def total_rms(self):
        return float(np.sqrt(np.mean(np.square(self, dtype=np.float32))))

    def moving_average(self, width: float):
        n = len(self)
        r = max(1, int(0.5 + width * self.rate / 2))
        assert r > 0 and n > 1
        idxlo = np.maximum(np.arange(n) - r, 0)
        idxhi = np.minimum(np.arange(n) + r, n - 1)
        cum = np.cumsum(self)
        mav = (cum[idxhi] - cum[idxlo]) / (idxhi - idxlo)
        return self.rated(mav)

    def envelope(self, width: float):
        rms = self.rated(np.square(self.astype(np.float32)))
        mav = np.sqrt(rms.moving_average(width))
        return self.rated(mav)

    @cached_property
    def _cached_std_envelope(self):
        env = self.envelope(0.5)
        #env = env.moving_average(3)
        return self.rated(env / env.std())

    @cached_method(maxsize=5)
    def cached_resample(self, rate: float):
        return self.resample(rate)

    @cached_property
    def time(self):
        return np.arange(len(self)) / self.rate

    def plot(self, ax=None):
        import matplotlib.pyplot as plt
        if ax == None:
            ax = plt.gca()
        return ax.plot(self.time, self)

    @classmethod
    def from_video(cls, file, rate, mono=True):
        tmp = f'{file}.tmp.wav'
        try:
            out = ffmpeg.input(file)['a:0']
            out = ffmpeg.output(out, tmp, ar=rate, ac=int(mono))
            ffmpeg_run(out)
            seq = cls.from_wav(tmp)
        finally:
            if os.path.exists(tmp):
                os.remove(tmp)
        return seq

    def normalized_to(self, ref, width: float):
        that = np.maximum(ref.envelope(width), 1e-3)
        this = np.maximum(self.envelope(width), 1e-3)
        return self.clip_cast(self * (that / this))

    def to_ogg(self, file):
        assert file.endswith('.ogg'), f'File {file} must end in .ogg'
        wav = f'{file}.wav'
        self.to_wav(wav)
        ffmpeg_run(ffmpeg.output(ffmpeg.input(wav), file))
        os.remove(wav)
        return

    def resample(self, out_rate: float):
        in_duration = self.duration
        out_duration = self.duration * out_rate / self.rate
        out = self.respeed(in_duration, out_duration)
        out.rate = out_rate
        return out

    def respeed(self, in_duration: float, out_duration: float):
        '''
        output signal has out_duration and same rate as self, but
        the content resembles self from 0 to in_duration (at different playback speed)
        '''
        assert 0 < min(
            in_duration,
            out_duration), f'Non sense: {in_duration}s {out_duration}s'
        in_samples = int(0.5 + in_duration * self.rate)
        out_samples = int(0.5 + out_duration * self.rate)
        if out_duration == in_duration:
            seq = self.rated(self[:in_samples])
        else:
            assert 0 < out_samples < 3e8, (out_samples, in_samples, in_duration,
                                           out_duration)
            idx = np.linspace(0, in_samples, out_samples, endpoint=False)
            seq = self._float_index(idx)
        return seq

    def _synced_to_plots(self, ref, start, end):
        this, that = self._synced_to_valid(ref, start, end)
        import matplotlib.pyplot as plt
        plt.plot(this.time, this)
        plt.plot(that.time, that, alpha=0.5)
        plt.show()
        #plt.scatter(this, that, alpha=0.1, marker='.')
        #plt.show()
        return

    def _synced_to_corr(self, ref, start, end):
        x, y = self._synced_to_valid(ref, start, end)
        C = np.corrcoef(x, y)
        total_duration = (self.duration + ref.duration)
        valid_duration = (x.duration + y.duration)
        return C[0, 1] * valid_duration / total_duration

    def _synced_to_valid(self, ref, start, end):
        'trims self and ref to the valid interval (no padding)'
        this, lo, hi = self._synced_to(ref, start, end)
        this = self.rated(this[:hi - lo])
        that = self.rated(ref[lo:hi])
        return this, that

    def _synced_to(self, ref, start, end):
        seq = self
        if seq.rate != ref.rate:
            seq = seq.cached_resample(ref.rate)
        in_duration = seq.duration
        out_duration = end - start
        seq = seq.respeed(in_duration, out_duration)
        lo = int(start * ref.rate)
        if lo < 0:
            seq = self.rated(seq[-lo:])
            lo = 0
        hi = min(int(end * ref.rate), lo + len(seq), len(ref))
        return seq, lo, hi

    def synced_to(self, ref, start, end, padcopy=True, width_normalize=10,
                  width_softpad=2, width_softstart=0.2):
        '''
        sync self to ref.
        out starts at start and ends at end (seconds).
        out is stretched if neccesary.
        out is trimmed and padded to fill (0, self.duration) 
        out is padded with a copy of ref if padcopy==True
        out padding is soften around border
        '''
        seq, lo, hi = self._synced_to(ref, start, end)
        out = self.rated(np.zeros_like(ref))
        out[lo:hi] = seq[:hi - lo]
        if width_normalize > 0:
            out = out.normalized_to(ref, width_normalize)
        if padcopy:
            out[:lo] = ref[:lo]
            out[hi:] = ref[hi:]
        if width_softpad > 0:
            k = min(int(ref.rate * width_softpad), hi - lo)
            p = np.linspace(0, 1, k)
            out[lo:lo + k] = p * out[lo:lo + k] + (1 - p) * ref[lo:lo + k]
            out[hi - k:hi] = (1 - p) * out[hi - k:hi] + p * ref[hi - k:hi]
        if width_softstart > 0:
            k = min(int(ref.rate * width_softstart), len(out) // 2)
            p = np.linspace(0, 1, k)
            out[:k] = p * out[:k]
            out[-k:] = (1 - p) * out[-k:]
        return out

    @cached_property
    def _std_envelope(self):
        env = self.envelope(5)
        env /= self.envelope(60)
        env = (env / env.mean() - 1) / env.std()
        return self.rated(env)

    def _sync_to_delays(self, ref):
        start = (ref.duration - self.duration) / 2
        rend = -start
        d_ref = ref.duration
        d_self = self.duration
        this = _this = self._std_envelope.resample(200)
        that = _that = ref._std_envelope.resample(200)
        radius = max(3, 1.3 * abs(start))
        nparts = 31
        corr = 0
        a = start
        b = rend
        while radius > 0.001:
            rate = min(nparts / radius, 200)
            this = _this.resample(rate)
            that = _that.resample(rate)
            points = []
            search = radius * np.linspace(-1, 1, nparts)
            for a in search + start:
                for b in search + rend:
                    d = d_ref + b - a
                    drastic = d <= 0 or max(d / d_self, d_self / d) >= 1.2
                    if not drastic:
                        c = this._synced_to_corr(that, a, d_ref + b)
                        points.append((c, a, b))
            corr, start, rend = max(*points)
            radius *= min(max(2.5 / nparts, 0.25), 0.75)
            nparts = max(5, 1 + nparts // 2)
        if corr < 0.3:
            print(
                f'  r={radius:.3f} [{start:.3f}s, {rend:.3f}s] corr={corr:.4f}')
            this._synced_to_plots(that, a, d_ref + b)
        #assert corr >= 0.3, (f'Insufficient correlation', corr, start, rend)
        return start, d_ref + rend, corr



from __future__ import annotations
from collections import deque, Counter, OrderedDict
import json
from json import encoder
import sqlite3
from types import FunctionType
from typing import Any, Callable, Dict, Generic, List, Mapping, Optional, Sequence, Tuple, Type, TypeVar, Union

DictEncoder = Tuple[Type, Callable[[Any], Dict]]
ListEncoder = Tuple[Type, Callable[[Any], List]]

JSONLike = Union[str, int, float, bool, None,
                 object]  # (recursive type hotfix, May 2021)


class CustomJSON:

    default_list_encoders: List[ListEncoder] = [
        (tuple, list),
        (deque, list),
        (set, list),
        (frozenset, list),
    ]
    default_dict_encoders: List[DictEncoder] = [
        (Counter, dict),
        (OrderedDict, dict),
    ]

    json_types: Sequence[Type] = [dict, list, int, str, float, bool, type(None)]

    def __init__(self, list_encoders: List[ListEncoder] = None,
                 dict_encoders: List[DictEncoder] = None):
        'encoders is a dict [class]->converter, where class has a __name__ and'
        'converter converts class into a list or a dict'
        self.list_encoders = dict(
            [*self.default_list_encoders, *(list_encoders or [])])
        self.dict_encoders = dict(
            [*self.default_dict_encoders, *(dict_encoders or [])])

    def stringify(self, obj: object):
        return self._stringify(obj)[1]

    def _stringify(self, obj: object):
        pre_dumped = self.pre_dump(obj)
        return pre_dumped, json.dumps(obj)

    def parse(self, text: str):
        return self.post_load(json.loads(text))

    def save(self, fname: str, obj: object):
        with open(fname, 'w') as f:
            json.dump(self.pre_dump(obj), f)

    def load(self, fname: str):
        with open(fname, 'r') as f:
            obj: JSONLike = json.load(f)
        return self.post_load(obj)

    def pre_dump(self, obj: object) -> JSONLike:
        'convert obj into a JSON serializable object recursively'
        out = self._prefix_pre_dump(obj)
        if isinstance(out, dict):
            out = {k: self.pre_dump(v) for k, v in out.items()}
        elif isinstance(out, list):
            out = [self.pre_dump(v) for v in out]
        return out

    def _prefix_pre_dump(self, obj: object):
        'convert obj to a JSON serializable only on the first depth level'
        'add prefixes for storing decoding info'
        if isinstance(obj, dict):
            return {(f'**{k}' if k.startswith('*') else k): v
                    for k, v in obj.items()}
        for cls, to_js in self.list_encoders.items():
            if isinstance(obj, cls):
                return {'*': cls.__name__, '**': to_js(obj)}
        for cls, to_js in self.dict_encoders.items():
            if isinstance(obj, cls):
                return {'*': cls.__name__, **self._prefix_pre_dump(to_js(obj))}
        if any(isinstance(obj, cls) for cls in self.__class__.json_types):
            return obj
        raise Exception(f'Unknown class {type(obj)} of object:\n{obj}')

    def post_load(self, loaded: JSONLike) -> object:
        'convert loaded JSONLike into decoded classes recursively'
        out: object = loaded
        if isinstance(out, list):
            out = [self.post_load(v) for v in out]
        elif isinstance(out, dict):
            out = {k: self.post_load(v) for k, v in out.items()}
            out = self._prefix_post_load(out)
        return out

    def _prefix_post_load(self, loaded: dict[str, JSONLike]):
        out: object = loaded
        if '*' in out:
            clsname = out.pop('*')
            cls = self.decoders[clsname]
            out = cls(out.pop('**'))
        # Undo dumped ficticious prefixes
        if isinstance(out, dict):
            out = {
                (k[2:] if k.startswith('**') else k): v for k, v in out.items()
            }
        return out

    @property
    def decoders(self):
        decoders = {
            **{cls.__name__: cls for cls in self.list_encoders},
            **{cls.__name__: cls for cls in self.dict_encoders}
        }
        return decoders


class DiskDict:
    '''
    Holds a dictionary in disk (sqlite3) using a CustomJSON class for conversion
      Keys must be text.
      Values must support serialization with JSON.parse
    
    If values are dictionaries, some queries on their keys are supported.
    
    #ToDo: Strings are forcibly unidecoded and lowercased to facilitate indexing
    '''

    def __init__(self, file=':memory:', JSON=None, indices=None):
        self.connection = sqlite3.connect(file)

        self.connection.execute(f'''
      CREATE TABLE IF NOT EXISTS objects(
        key TEXT NOT NULL PRIMARY KEY,
        obj TEXT NOT NULL)''')
        self.connection.execute('''
      CREATE TABLE IF NOT EXISTS indexed(
        key TEXT NOT NULL)''')
        self.JSON = JSON or CustomJSON()
        self.add_indices(indices or [])

    def _execommit(self, *query):
        try:
            self.connection.execute(*query)
        except Exception as e:
            raise Exception(*query) from e
        self.connection.commit()

    def _iterrows(self, *query):
        yield from self.connection.execute(*query)

    def _itercol(self, *query):
        it = (t for t, *_ in self._iterrows(*query))
        yield from it

    def __iter__(self):
        q = (f'SELECT key FROM objects',)
        yield from self._itercol(*q)

    def __contains__(self, key):
        q = (f'SELECT key FROM objects WHERE key=?', (key,))
        yield any(k == key for k in self._itercol(*q))

    def __getitem__(self, key):
        q = (f'SELECT obj FROM objects WHERE key=?', (key,))
        found = list(map(self.JSON.parse, self._itercol(*q)))
        if not found:
            raise KeyError(key)
        return found.pop()

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def items(self):
        q = (f'SELECT key, obj FROM objects',)
        for key, obj in self.connection.execute(*q):
            yield key, self.JSON.parse(obj)

    def values(self):
        q = (f'SELECT obj FROM objects',)
        yield from map(self.JSON.parse, self._itercol(*q))

    def __setitem__(self, key, obj):
        pre_dumpd, value = self.JSON._stringify(obj)
        if key in self:
            self._del_indexed(key)
        self._insert_indexed(((key, pre_dumpd),))
        self._execommit(f'REPLACE INTO objects VALUES (?,?)', (key, value))

    @property
    def indexed(self):
        value = self.__dict__.get('__indexed')
        if value is None:
            value = self._table_columns('indexed')[1:]
            self.__dict__['__indexed'] = value
        return value

    def _insert_indexed(self, items, table='indexed'):
        if table == 'indexed':
            indexed = self.indexed
        else:
            indexed = self._table_columns(table)[1:]
        qmark = ','.join('?' for _ in range(len(indexed) + 1))
        query = f'INSERT INTO {table} VALUES ({qmark})'
        for key, obj in items:
            if hasattr(obj, 'get'):
                values = (obj.get(key) for key in indexed)
                self._execommit(query, (key, *values))

    def Keys(self):
        return list(self)

    def Values(self):
        return list(self.values())

    def Items(self):
        return list(self.items())

    def Dict(self):
        return dict(self.items())

    def indexed_items(self):
        yield from self.connection.execute(f'SELECT * FROM indexed')

    def where_query(self, params, limit: Optional[int]):
        'Usage: see self.where'
        indexed = self.indexed
        values = []

        def parse(column, eq, value):
            nonlocal values, indexed
            if eq.lower() in ('and', 'or'):
                return f'({parse(*column)} {eq} {parse(*value)})'
            assert eq in ('==', '!=', '<', '<=', '>',
                          '>='), f'Invalid operator: {eq}'
            assert column in indexed, f'Can not run query on non-indexed column "{column}"'
            if value is None:
                assert eq in ('==', '!='), f'Invalid operator with NULL: {eq}'
                if eq == '==':
                    return f'({column} IS NULL)'
                return f'({column} IS NOT NULL)'
            if eq == '!=':
                eq = '<>'
            elif eq == '==':
                eq = '='
            values.append(value)
            return f'({column} {eq} ?)'

        qlimit = '' if limit is None else f'LIMIT {limit}'
        return (f'SELECT key FROM indexed WHERE {parse(*params)} {qlimit}',
                values)

    def where(self, params, limit=None):
        '''
    Usage:
      where((column, '<', value), limit=10)
      where(((('n', '==', 3), 'and', ('k', '>=', 4)), 'or' ('name', '!=', 'carlos')))
    '''
        q = self.where_query(params, limit=limit)
        yield from (self[key] for key in self._itercol(*q))

    def __delitem__(self, key):
        self._execommit(f'DELETE FROM objects WHERE key=?', (key,))
        self._del_indexed(key)

    def _del_indexed(self, key):
        self._execommit(f'DELETE FROM indexed WHERE key=?', (key,))

    def del_index(self, column):
        assert column != 'key'
        self._execommit(
            f'CREATE TABLE tmp_indexed AS SELECT * FROM indexed WHERE 1=0')
        self._execommit(f'ALTER TABLE tmp_indexed DROP COLUMN {column}')
        self._insert_indexed(self.items(), table='tmp_indexed')
        self._execommit(f'ALTER TABLE indexed RENAME TO tmp__indexed')
        self._execommit(f'ALTER TABLE tmp_indexed RENAME TO indexed')
        self._execommit(f'DROP TABLE IF EXISTS tmp__indexed')

    def _table_columns(self, table):
        q = (f'SELECT name FROM PRAGMA_TABLE_INFO(?)', (table,))
        return list(self._itercol(*q))

    def __len__(self):
        return next(self._itercol(f'SELECT COUNT(key) FROM objects'))

    def add_indices(self, indices):
        for column, sql_type in indices:
            self.add_index(column, sql_type)

    def add_index(self, column, sql_type):
        if column not in self.indexed:
            self._add_index(column, sql_type)

    def _add_index(self, column, sql_type):
        self._execommit(f'DROP TABLE IF EXISTS tmp_indexed')
        self._execommit(
            f'CREATE TABLE tmp_indexed AS SELECT * FROM indexed WHERE 1=0')
        self._execommit(
            f'ALTER TABLE tmp_indexed ADD COLUMN {column} {sql_type}')
        self._insert_indexed(self.items(), table='tmp_indexed')
        self._execommit(f'ALTER TABLE indexed RENAME TO tmp__indexed')
        self._execommit(f'ALTER TABLE tmp_indexed RENAME TO indexed')
        self._execommit(f'DROP TABLE IF EXISTS tmp__indexed')
        self.__dict__.pop('__indexed', None)



import collections.abc
from pprint import PrettyPrinter
from typing import Any, List, Tuple

pprinter = PrettyPrinter()


class Dict(dict):

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            return object.__getattribute__(self, key)

    def __setattr__(self, key, value):
        self[key] = value

    @staticmethod
    def recursive(d: dict):
        root = Dict(d)
        for k, v in root.items():
            if isinstance(v, dict) and (not isinstance(v, Dict)):
                root[k] = Dict.recursive(v)
        return root

    def __str__(self):
        return pprinter.pformat(self)


def deep_update(d, u, _class=dict):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = deep_update(d.get(k, _class()), v)
        else:
            d[k] = v
    return d


def flatten(d: dict) -> dict:
    flat = {}
    q: List[Tuple[str, Any]] = [('', d)]
    while q:
        prefix, e = q.pop()
        if isinstance(e, dict):
            for key, val in e.items():
                q.append((f'{prefix}.{key}', val))
        else:
            flat[prefix[1:]] = e
    return flat


def unflatten(flat: dict) -> dict:
    root = {}
    for key, val in flat.items():
        parent = root
        for sub in key.split('.'):
            parent[sub] = parent.get(sub, {})
            last = parent
            parent = parent[sub]
        last[sub] = val
    return root



from __future__ import annotations
from collections import deque, Counter, OrderedDict
from contextlib import contextmanager
import json, time, random, logging
from json import encoder
import sqlite3
from types import FunctionType
from typing import Any, Callable, Dict, Generic, Iterable, List, Literal, Mapping, Optional, Tuple, Type, TypeVar, Union


def custom_repr(self, *keys):
    name = self.__class__.__name__
    comma_kwargs = ', '.join(f'{k}={repr(getattr(self, k))}' for k in keys)
    return f'{name}({comma_kwargs})'


#Params = Excluding[str, Sequence[Any]] # If Excluding existed
#Columns = Excluding[str, Sequence[str]] # If Excluding existed
Params = Union[List[Any], Tuple[Any, ...]]
Columns = Union[List[str], Tuple[str, ...]]


class SQLiteDB:
    '''
    '''

    def __repr__(self):
        return custom_repr(self, 'file')

    def __init__(self, file=':memory:'):
        self.file = file

    def new_connection(self):
        return sqlite3.connect(self.file, check_same_thread=False)

    def _execute(self, query: str, params: Params = None) -> sqlite3.Cursor:
        assert not isinstance(params, str), f'Did you mean params=[{params}]?'
        try:
            with self.new_connection() as con:
                return con.execute(query, params or [])
        except sqlite3.OperationalError as e:
            e.args = (*e.args, query, str(params))
            raise e

    def execute(self, query: str, params: Params = None) -> List[Any]:
        return [*self._execute(query, params or [])]

    def execute_column(self, query: str, params: Params = None):
        rows = self.execute(query, params)
        return [first for first, *_ in rows]

    _query_table_names = """
        SELECT name FROM sqlite_master 
        WHERE type = 'table' 
        AND name NOT LIKE 'sqlite_%'
        ORDER BY 1;
    """

    def table_names(self):
        return self.execute_column(self._query_table_names)

    _query_index_names = """
        SELECT name FROM sqlite_master 
        WHERE type = 'index' 
        ORDER BY 1;
    """

    def index_names(self):
        return self.execute_column(self._query_index_names)

    def tables(self):
        names = self.table_names()
        return {t: self.get_table(t) for t in names}

    def get_table(self, table_name: str):
        return SQLiteTable(self.file, table_name)

    def get_indexed_table(
        self,
        table_name: str,
        index_columns: List[str],
    ):
        table = self.get_table(table_name)
        return table.indexed_by(index_columns)

    def get_key_value_table(self, table_name: str):
        return self.get_table(table_name).as_key_value()

    def table_columns(self, table_name: str):
        return self.get_table(table_name).columns()

    def table_count(self, table_name: str):
        return len(self.get_table(table_name))


class SQLiteTable:

    def __init__(self, file: str, table_name: str):
        self.file = file
        self.table_name = table_name
        self.db = SQLiteDB(self.file)

    def __repr__(self):
        return custom_repr(self, 'file', 'table_name')

    _query_columns = """
    SELECT name FROM PRAGMA_TABLE_INFO(?)
    """

    def columns(self):
        return self.db.execute_column(self._query_columns, [self.table_name])

    def __len__(self) -> int:
        query = f'SELECT COUNT(*) FROM {self.table_name}'
        return self.db.execute(query)[0][0]

    def insert_row(self, **kwargs):
        assert kwargs, 'Nothing to set'
        table = self.table_name
        columns = ', '.join(kwargs.keys())
        marks = ', '.join('?' for _ in kwargs)
        query = f'INSERT INTO {table} ({columns}) VALUES ({marks})'
        params = [*kwargs.values()]
        return self.db._execute(query, params)

    def _make_query(self, columns: Columns = None, where: str = None):
        table = self.table_name
        what = self._make_columns(columns)
        query = f'SELECT {what} FROM {table}'
        if where is not None:
            query = f'{query} WHERE {where}'
        return query

    def _make_columns(self, columns: Columns = None):
        if columns is None:
            comma_columns = '*'
        else:
            # Security check:
            assert set(columns) <= set(self.columns())
            comma_columns = ','.join(columns)
        return comma_columns

    def rows(self, columns: Columns = None, where: str = None,
             params: Params = None):
        query = self._make_query(columns, where)
        return self.db.execute(query, params)

    def _to_dict_map(self, columns: Columns = None):
        columns = self.columns() if not columns else columns
        to_dict = lambda row: dict(zip(columns, row))
        return to_dict

    def dicts(self, columns: Columns = None, where: str = None,
              params: Params = None):
        rows = self.rows(columns, where, params)
        to_dict = self._to_dict_map(columns)
        return [to_dict(row) for row in rows]

    def column(self, field: str = None, where: str = None,
               params: Params = None):
        columns = None if field is None else [field]
        return [r[0] for r in self.rows(columns, where, params)]

    def unique_dict(self, columns: Columns = None, where: str = None,
                    params: Params = None):
        dicts = self.dicts(columns, where, params)
        n = len(dicts)
        assert n <= 1, f'Multiple ({n}) results, e.g.: {dicts[:2]}'
        return dicts[0] if dicts else None

    def get_first_dict(self, columns: Columns = None):
        return self.unique_dict(columns, '1=1 LIMIT 1', None)

    def indexed_by(self, index_columns: List[str]):
        return IndexedSQLiteTable(
            self.file,
            self.table_name,
            index_columns,
        )

    def as_key_value(self):
        return KeyValueSQLiteTable(self.file, self.table_name)


class IndexedSQLiteTable(SQLiteTable):

    _repr_keys = ['file', 'table_name', 'index_columns']

    def __init__(self, file: str, table_name: str, index_columns: List[str]):
        super().__init__(file, table_name)
        self.index_columns = index_columns

    def __repr__(self):
        return custom_repr(self, 'file', 'table_name', 'index_columns')

    def _where(self):
        columns = self.index_columns
        return ' AND '.join(f'{c}=?' for c in columns)

    def get_row(self, *idx: Any, columns: Columns = None):
        assert idx, 'Nowhere to get'
        where = self._where()
        rows = self.rows(columns, where, idx)
        return rows[0] if rows else None

    def get_dict(self, *idx: Any, columns: Columns = None):
        assert idx, 'Nowhere to get'
        where = self._where()
        dicts = self.dicts(columns, where, idx)
        return dicts[0] if dicts else None

    def update_row(self, *idx: Any, **kwargs):
        assert idx, 'Nowhere to set'
        assert kwargs, 'Nothing to set'
        table = self.table_name
        what = ', '.join(f'{c}=?' for c in kwargs.keys())
        where = self._where()
        query = f'UPDATE {table} SET {what} WHERE {where}'
        params = [*kwargs.values(), *idx]
        cur = self.db._execute(query, params)
        did_update = (cur.rowcount > 0)
        return did_update

    def set_row(self, *idx: Any, **kwargs):
        assert idx, 'Nowhere to set'
        for c, v in zip(self.index_columns, idx):
            if c in kwargs and kwargs[c] != v:
                msg = (f'Inconsistent idx and kwargs at column ({c}):'
                       f' ({v}) vs ({kwargs[c]})')
                raise Exception(msg)
            kwargs[c] = v
        if not self.update_row(idx, **kwargs):
            self.insert_row(**kwargs)
        return

    def del_row(self, *idx: Any):
        assert idx, 'Nowhere to del'
        table = self.table_name
        where = self._where()
        query = f'DELETE FROM {table} WHERE {where}'
        cur = self.db._execute(query, idx)
        did_del = cur.rowcount
        return did_del


class KeyValueSQLiteTable(SQLiteTable):

    def __init__(self, file: str, table_name: str):
        super().__init__(file, table_name)
        self.db.execute(f'''
        CREATE TABLE IF NOT EXISTS {self.table_name}(
            key text NOT NULL PRIMARY KEY,
            value text NOT NULL,
            lock_token double NOT NULL,
            locked_until double NOT NULL
        )
        ''')
        assert set(self.columns()) == {'key', 'value', 'lock'}
        return

    @contextmanager
    def exclusive_access(self, *keys: str, delta=0.02, max_duration=0.5):
        assert keys, 'You must specify keys to be locked explicitely'
        token = 1 + random.random()
        for key in keys:
            self._lock(key, token, delta, max_duration)
        try:
            yield
        finally:
            for key in keys:
                self._unlock(key, token, max_duration)
        return

    def _current_lock(self, key: str):
        lock_keys = ['lock_token', 'locked_until']
        return super().unique_dict(lock_keys, 'key=?', [key])

    def _lock(self, key: str, token: float, delta=0.02, max_duration=1.5):
        while True:
            d = self._current_lock(key)
            if not d:
                raise KeyError(key)
            if d['lock_token'] == 0 or time.time() > d['locked_until']:
                # Request access, race until next loop
                super().indexed_by(['key']).set_row(
                    key,
                    lock_token=token,
                    locked_until=time.time() + max_duration,
                )
            elif d['lock_token'] == token:
                # Access gained
                break
            time.sleep(delta)
        return token

    def _unlock(self, key: str, token: float, max_duration: float):
        d = self._current_lock(key)
        if not d:
            logging.warning(
                f'Key {repr(key)} was deleted by another thread or process during exclusive access'
            )
            return
        remaining = d['locked_until'] - time.time()
        if remaining < 0:
            ms = round(-remaining * 1000)
            logging.warning(
                f'Locked {repr(key)} during {ms}ms more than max_duration={max_duration}s'
            )
        if d['lock_token'] == token:
            super().indexed_by(['key']).set_row(key, lock_token=0)
        return

    def __getitem__(self, key: str):
        d = super().unique_dict(['value'], 'key=?', [key])
        if not d:
            raise KeyError(key)
        return json.loads(d['value'])

    def __setitem__(self, key: str, value: Any):
        super().indexed_by(['key']).set_row(key, value=json.dumps(value))
        return

    def values(self):
        return [json.loads(s) for s in super().column('value')]

    def keys(self) -> List[str]:
        return [k for k in super().column('key')]

    def items(self) -> List[Tuple[str, Any]]:
        items = super().rows(['key', 'value'])
        return [(k, json.loads(v)) for k, v in items]

    def get(self, key: str, default=None):
        try:
            return self[key]
        except KeyError:
            return default


def test():
    db = SQLiteDB('.test.db')
    db.execute(f'''
    CREATE TABLE IF NOT EXISTS objects(
        key TEXT NOT NULL PRIMARY KEY,
        obj TEXT NOT NULL
    )''')
    db.execute('''
    CREATE TABLE IF NOT EXISTS indexed(
        key TEXT NOT NULL
    )''')
    #print(db.tables())
    #print(db.index_names())
    #print(db.execute('SELECT * FROM sqlite_master'))
    ob = db.get_indexed_table('objects', ['key'])
    print(ob)
    print(len(ob))
    ob.set_row('key1', obj='VALUE1')
    print(ob.get_row('key1'))
    print(len(ob))


from typing import Union, Dict
from unidecode import unidecode
import re, json, time, requests, mimetypes
from mutagen.id3 import ID3NoHeaderError, Encoding
from mutagen.id3 import ID3, TIT2, TALB, TPE1, TPE2, COMM, TCOM, TCON, TDRC, TRCK, USLT, APIC, TXXX
import lyricsgenius

def is_latin(s:str):
    is_latin = True
    try:
        s.encode('latin1')
    except:
        is_latin = False
    return is_latin


def latinize(s:str):
    '''(non-performant) unidecode for latin1 instead of ascii'''
    if is_latin(s):
        return s
    fixed = ''.join(c if is_latin(c) else unidecode(c) for c in s)
    assert is_latin(fixed)
    return fixed

genius_token = 'SOME TOKEN'
genius = lyricsgenius.Genius(genius_token)


class MyID3(ID3):
    
    _TAGS = dict(
        artist=TPE1,
        title=TIT2,
        album=TALB,
        track=TRCK,
        year=TDRC,
        comments=COMM,
        lyrics=USLT,
        picture=APIC,
        genius_id=TXXX,
        youtube_id=TXXX,
    )
    
    def __init__(self, fname:str):
        try:
            super().__init__(fname)
        except ID3NoHeaderError:
            super().__init__()
            self.save(fname)
        
        self.latinize_tags()
        self.make_lyrics_unique()
        self.make_picture_unique()
        # Remove garbage
        if str(self.get_tag('album', '')).lower() == 'youtube downloader audio':
            self.write_tag('album', text='')
    
    @classmethod
    def _TAG(cls, tag_name:str):
        tag_cls = cls._TAGS.get(tag_name, TXXX)
        tag_key = tag_cls.__name__
        
        if tag_cls is TXXX: # Custom tag
            kwargs = dict(
                encoding=Encoding.LATIN1,
                desc=tag_name,
                text='',
            )
            tag_key = f'TXXX:{tag_name}'
        elif tag_cls is APIC: # Picture tag
            kwargs = dict(
                encoding=Encoding.UTF8,
                data=bytes(),
                mime='image/jpeg',
                type=3,
                desc='Cover',
            )
        elif tag_cls is USLT: # Lyrics tag
            kwargs = dict(
                encoding=Encoding.LATIN1,
                text='',
                desc='',
                lang='XXX',
            )
        else: # Standard (text) tag
            kwargs = dict(
                encoding=Encoding.LATIN1,
                text='',
            )
        return tag_key, tag_cls, kwargs
    
    @classmethod
    def _LATIN_TAGS(cls):
        '''Tags that should be encoded using LATIN1'''
        zipped = [cls._TAG(key) for key in cls._TAGS]
        return [(k,cl,kw) for k,cl,kw in zipped if kw['encoding']==Encoding.LATIN1]
    
    def latinize_tags(self):
        for (k,v,kw) in self._LATIN_TAGS():
            value = self.get(str(v))
            if value and value.encoding != Encoding.LATIN1:
                value.text = latinize(value.text)
                value.encoding = Encoding.LATIN1
                self.save()
        return
        
    def make_lyrics_unique(self):
        '''
        Force at most one lyrics USLT tag
        Force also the default values
        '''
        uslt = [key for key in self.keys() if key.startswith('USLT')]
        lyrics = max([self[key].text for key in uslt]+[''], key=lambda s: len(str(s)))
        self.delall('USLT')
        self.write_tag('lyrics', text=lyrics)
        
    def make_picture_unique(self):
        '''
        Force at most one APIC tag
        Force also desc=''
        '''
        apic = [key for key in self.keys() if key.startswith('APIC')]
        if apic:
            best = self[max(apic, key=lambda key: len(str(self[key].data)))]
            self.delall('APIC')
            self.write_tag('picture', data=best.data, mime=best.mime, encoding=best.encoding)
    
    
    def write_tag(self, tag_name, **kwargs):
        tag_key, tag_cls, tag_kwargs = self._TAG(tag_name)
        kwargs = {**tag_kwargs, **kwargs}
        
        if kwargs['encoding'] == Encoding.LATIN1 and 'text' in kwargs:
            kwargs['text'] = latinize(kwargs['text'])
        
        self[tag_key] = tag_cls(**kwargs)
        self.save()
        
    def read_tag(self, tag_name):
        tag_key, _, _ = self._TAG(tag_name)
        return self[tag_key]
    
    def get_tag(self, tag_name, default=None):
        tag_key, _, _ = self._TAG(tag_name)
        return self.get(tag_key, default)
    
    def get_json(self, tag_name, default=None):
        tag_key, _, _ = self._TAG(tag_name)
        tag = self.get(tag_key, None)
        if tag:
            try:
                data = json.loads(str(tag))
            except:
                data = default
        else:
            data = default
        return data
    
    def __str__(self):
        title = str(self.read_tag('title'))
        artist = str(self.read_tag('artist'))
        tags = ['genius_json', 'lyrics', 'picture']
        plus = ''.join(f'+{tag}' for tag in tags if self.get_tag(tag))
        return f'"{title}" by {artist} [{plus}]'


def auto_lyrics(genius, id3, recent_days=30, sleep_seconds=5):
    last = id3.get_json('genius_timestamp')
    searched_recently = False
    days = 3600 * 24
    if isinstance(last, float):
        searched_recently = (time.time() < last + recent_days*days)
    
    data = id3.get_json('genius_json')
    if not data and not searched_recently:
        title = str(id3.read_tag('title'))
        artist = str(id3.read_tag('artist'))
        time.sleep(sleep_seconds)
        response = genius.search_song(title, artist) # (None if not found)
        id3.write_tag('genius_timestamp', text=json.dumps(time.time()))
        if response:
            id3.write_tag('genius_json', text=response.to_json())
            data = id3.get_json('genius_json', {})
            lyrics = data.get('lyrics', '')
            id3.write_tag('lyrics', text=lyrics)
    return


def auto_picture(id3, sleep_seconds=3):
    if not id3.get_tag('picture'):
        url = id3.get_json('genius_json', {}).get('song_art_image_url', '')
        mime, _ = mimetypes.guess_type(url)
        if url and mime:
            time.sleep(sleep_seconds)
            data = requests.get(url, stream=True).raw.data
            id3.write_tag('picture', data=data, mime=mime)
    return


import os

def print_enum(it):
    l = list(it)
    yield from ((print(f'{i}/{len(l)}') or x) for i,x in enumerate(l))
    print(f'({len(l)}/{len(l)})')

def del_all_pictures(id3s):
    for id3 in id3s:
        id3.delall('APIC')
        id3.save()
    return


xdir = '/home/carlos/Mega/Redmi 8/songs/Instrumental A/'
id3s = [MyID3(f'{xdir}/{xfile}') for xfile in sorted(os.listdir(xdir))]

for id3 in print_enum(id3s):
    print(id3)
    try:
        auto_lyrics(genius, id3)
        auto_picture(id3)
    except Exception as e:
        print(e)
    print(id3)




from typing import Any, Optional, TypedDict, Union
from ._dict import Dict
from subprocess import (Popen, PIPE, DEVNULL, TimeoutExpired,
                        CalledProcessError)
from tempfile import NamedTemporaryFile
from threading import Thread, Timer
from queue import Queue, Empty
import sys, time, io, asyncio, threading
from .interrupt import terminate_thread

ON_POSIX = 'posix' in sys.builtin_module_names


def _silent_interrupt(func):

    def wrapper(self, *args, **kwargs):
        try:
            func(self, *args, **kwargs)
        except KeyboardInterrupt:
            pass
        return

    wrapper.__name__ = func.__name__
    return wrapper


class MyProcess():
    '''
    Tool for launching a process with support for live stdout/stderr
    handling, as well as timeout, custom input and stdout/stderr
    capturing/non-caputring.
    Uses threads and queues.

    Arguments:
        args: arguments passed to Popen
        shell: passed to Popen
        input: string that simulates stdin
        timeout: None or float timeout in seconds
        check: bool. If true and an error ocurred, throws an exception
        capture_stdout: bool. If true, output is assigned to 'self.stdout'
        live_stdout: None, io.BufferedWriter instance, or any object
            with methods write and flush. If given, write(line) and flush()
            are called for each line of stdout as soon as it is received
            (as soon as the program flushes), with a maximum delay of
            'micro_delay'.
        timeout_stdout: None (infinity) or float. Time in seconds to
            wait for the live_stdout handler after the process terminates.
            Recall that the live_stdout handler may still have some
            remaining lines to handle once the process finishes.
            The additional time will not affect 'self.elapsed'.
        max_stdout: None, int or string. If more than max_stdout bytes
            are received, the process is killed. Notice that if capturing
            and live_stdout are disabled, no bytes will be received at all.
            Accepts string of the form '#', '#k', '#M', '#G', '#T'.
        capture_stderr: (see capture_stdout)
        stderr_handler: (see stdout_handler)
        wait_stderr_handler: (see wait_stdout_handler)
        max_stderr: (see max_stdout)
        encoding: string used for encoding/decoding stdin/stdout/stderr
        micro_delay: float seconds (see stdout_handler)
        block: bool. Blocks the current thread unitl the process finishes.
            If false, you must call wait()

    Returns None but sets:
        self.stdout: string, stdout of the process (if captured)
        self.stderr: string, stderr of the process (if captured)
        self.elapsed: float, approximate elapsed time of the process
            in seconds.
        self.timeout: float, copy of the timeout argument
        self.error: None or string, either 'TimeoutExpired',
            'ExcessOfOutput', 'KeyboardInterrupt',
            'NonZeroExitCode #', or an unexpected exception as string.
        self.returncode: int, exit code of the process
    '''

    def __init__(self, args, shell=False, env=None, cwd=None):
        kwargs = locals()
        kwargs.pop('self')
        self.kwargs = Dict(kwargs)

    def run(
        self,
        input: str = None,
        timeout: float = None,
        check: bool = False,
        capture_stdout=True,
        live_stdout=None,
        timeout_stdout=None,
        max_stdout=None,
        capture_stderr=False,
        live_stderr=None,
        timeout_stderr=None,
        max_stderr=None,
        encoding='utf-8',
        micro_delay=1e-3,
    ):
        kwargs = locals()
        kwargs.pop('self')
        self.kwargs.update(kwargs)
        self._start()
        interrupt = Thread(
            target=self._kill,
            args=['KeyboardInterrupt'],
        )
        try:
            #self._sync.wait() was here before but
            #the loop is needed for quick handling of
            #terminate_thread(thread, KeyboardInterrupt)
            while not self._sync.is_set():
                time.sleep(1e-4)
        except KeyboardInterrupt:
            interrupt.start()
        except Exception as e:
            self._error = str(e)
        while 1:
            try:
                self._sync.wait()
                if check and self.error:
                    raise CalledProcessError(
                        returncode=self.returncode,
                        cmd=self.kwargs.args,
                    )
                break
            except KeyboardInterrupt:
                pass
        return self

    async def async_run(self, input=None, timeout=None, check=False,
                        capture_stdout=True, live_stdout=None,
                        timeout_stdout=None, max_stdout=None,
                        capture_stderr=False, live_stderr=None,
                        timeout_stderr=None, max_stderr=None, encoding='utf-8',
                        micro_delay=1e-3):
        kwargs = locals()
        kwargs.pop('self')
        self.kwargs.update(kwargs)
        self._start()
        assert self._async
        await self._async.wait()
        if check and self.error:
            raise CalledProcessError(
                returncode=self.returncode,
                cmd=self.kwargs.args,
            )
        return self

    def run_detached(
        self,
        input: str = None,
        timeout: float = None,
        check: bool = False,
        capture_stdout=True,
        live_stdout=None,
        timeout_stdout=None,
        max_stdout=None,
        capture_stderr=False,
        live_stderr=None,
        timeout_stderr=None,
        max_stderr=None,
        encoding='utf-8',
        micro_delay=1e-3,
    ):
        kwargs = locals()
        kwargs.pop('self')
        self.kwargs.update(kwargs)
        self._start()
        return self

    def _start(self):
        self._done = False
        self._stop = False
        self._error = None
        self._timeout = self._parse_time(self.kwargs.timeout, None)
        self._micro_delay = self.kwargs.micro_delay
        self._encoding = self.kwargs.encoding
        self._max_stdout = self._parse_eng(self.kwargs.max_stdout)
        self._max_stderr = self._parse_eng(self.kwargs.max_stderr)
        self._timeout_stdout = self._parse_time(self.kwargs.timeout_stdout,
                                                float('inf'))
        self._timeout_stderr = self._parse_time(self.kwargs.timeout_stderr,
                                                float('inf'))
        self._threads = {}
        self._check = self.kwargs.check
        self._sync = threading.Event()
        try:
            self._async = asyncio.Event()
        except RuntimeError:
            self._async = None

        self._threads['waiter'] = Thread(target=self._waiter)
        if self._timeout:
            self._threads['timer'] = Timer(
                self._timeout,
                self._kill,
                args=['TimeoutExpired'],
            )

        config = {
            'out': {
                'capture': self.kwargs.capture_stdout,
                'live': self.kwargs.live_stdout,
                'wait': self._timeout_stdout,
                'max_size': self._max_stdout,
            },
            'err': {
                'capture': self.kwargs.capture_stderr,
                'live': self.kwargs.live_stderr,
                'wait': self._timeout_stderr,
                'max_size': self._max_stderr,
            },
        }

        for key, val in config.items():
            piped = val['capture'] or val['live']
            val['pipe'] = PIPE if piped else DEVNULL

        self._start_time = time.time()
        try:
            self._process = Popen(
                self.kwargs.args,
                shell=self.kwargs.shell,
                stdin=PIPE,
                stdout=config['out']['pipe'],
                stderr=config['err']['pipe'],
                #bufsize=1,
                close_fds=ON_POSIX,
                env=self.kwargs.env,
                cwd=self.kwargs.cwd,
            )
        except FileNotFoundError as e:
            self._stop = True
            self._done = True
            self._sync.set()
            if self._async:
                self._async.set()
            self.error = str(e)
            if self.kwargs.check:
                raise
            return
        assert self._process.stdin
        if self.kwargs.input != None:
            _input = self.kwargs.input.encode(self.kwargs.encoding)
            self._process.stdin.write(_input)
        self._process.stdin.close()

        config['out']['source'] = self._process.stdout
        config['err']['source'] = self._process.stderr

        self._buffer_stdout = io.StringIO(
        ) if config['out']['capture'] else None
        config['out']['buffer'] = self._buffer_stdout
        self._buffer_stderr = io.StringIO(
        ) if config['err']['capture'] else None
        config['err']['buffer'] = self._buffer_stderr

        for key, val in config.items():
            queues = []
            h = {}
            if val['capture']:
                h['capture'] = dict(ostream=val['buffer'], flush=False,
                                    wait=float('inf'))
            if val['live']:
                h['handler'] = dict(ostream=val['live'], flush=True,
                                    wait=val['wait'])
            for name, kwargs in h.items():
                queues.append(Queue())
                self._threads[f'std{key}_{name}'] = Thread(
                    target=self._live_handler,
                    args=[queues[-1]],
                    kwargs=kwargs,
                )
            if queues:
                self._threads[f'{key}-main'] = Thread(
                    target=self._non_blocking_reader,
                    kwargs=dict(istream=val['source'], queues=queues,
                                max_size=val['max_size']))

        for key, t in self._threads.items():
            t.start()
        return

    def _waiter(self):
        try:
            while not self._stop:
                if self._process.poll() != None:
                    self._stop = True
                else:
                    time.sleep(self._micro_delay)

            if self._process.stdout:
                self._process.stdout.close()
            if self._process.stderr:
                self._process.stderr.close()

            def get_value(buffer):
                if buffer == None:
                    return None
                value = buffer.getvalue()
                buffer.close()
                return value

            self._end = time.time()
            if 'timer' in self._threads:
                self._threads['timer'].cancel()

            self.stdout = get_value(self._buffer_stdout)
            self.stderr = get_value(self._buffer_stderr)
            self.elapsed = self._end - self._start_time
            self.timeout = self._timeout
            self.returncode = self._process.wait()
            if self._error:
                self.error = self._error
            elif self.returncode != 0:
                self.error = f'NonZeroExitCode {self.returncode}'
            else:
                self.error = None
            for key, t in self._threads.items():
                if key != 'waiter':
                    t.join()
        finally:
            self._done = True
            self._sync.set()
            if self._async:
                self._async.set()
        return

    def kill(self):
        self._kill('KilledByUser')
        while not self._done:
            time.sleep(1e-3)
        return

    def _kill(self, error):
        if self.is_active():
            self._error = error
        if 'timer' in self._threads:
            self._threads['timer'].cancel()
        self._stop = True
        self._process.kill()
        for k, t in self._threads.items():
            if k != 'waiter' and k != 'timer':
                terminate_thread(t, KeyboardInterrupt)
        for k, t in self._threads.items():
            if k != 'waiter' and k != 'timer':
                t.join()

    @_silent_interrupt
    def _non_blocking_reader(self, istream, queues, max_size):
        #https://stackoverflow.com/a/4896288/3671939
        for line in iter(istream.readline, b''):
            max_size -= len(line)
            if max_size < 0:
                self._stop = True
                self._error = 'ExcessOfOutput'
            if self._stop:
                break
            line = line.decode(self._encoding)
            for q in queues:
                q.put(line)
        return istream.close()

    @_silent_interrupt
    def _live_handler(self, queue, ostream, flush, wait):
        waiting = False
        waiting_start = None
        while not self._stop or waiting:
            try:
                elem = queue.get(timeout=self._micro_delay)
            except Empty:
                waiting = False
            else:
                ostream.write(elem)
                if flush:
                    ostream.flush()
            if self._stop:
                if waiting_start == None:
                    waiting_start = time.time()
                    waiting = True
                if time.time() - waiting_start > wait:
                    waiting = False
        return

    def _parse_eng(self, x):
        units = {'k': 1e3, 'M': 1e6, 'G': 1e9, 'T': 1e12}
        if x == None:
            return float('inf')
        elif isinstance(x, int):
            return x
        elif isinstance(x, float):
            return round(x)
        elif x.isdigit():
            return int(x)
        else:
            return int(x[:-1]) * int(units[x[-1]])

    def _parse_time(self, x, ifNone):
        return ifNone if x == None else x

    def is_active(self):
        return self._done == False


class Tee(io.BufferedWriter):
    """
    Simple BufferedWriter that broadcasts
    data to multiple BufferedWriters
    """

    def __init__(self, *outputs):
        self.outputs = outputs

    def write(self, s):
        for out in self.outputs:
            out.write(s)

    def flush(self):
        for out in self.outputs:
            out.flush()


class CustomOStream(io.BufferedWriter):

    def __init__(self, write_function, flush_function=None):
        self.write = write_function
        self.flush = flush_function or (lambda: 0)  # type:ignore
        self.tmp = NamedTemporaryFile('r', suffix='.out')
        self.fileno = self.tmp.fileno  # Provide a dummy fileno

    def __del__(self):
        self.tmp.close()
        object.__del__(self)  # type:ignore


class TemporaryStdout(io.RawIOBase):
    '''
    Replace stdout temporarily with another stream
    '''

    def __enter__(self):
        self.prev = sys.stdout
        sys.stdout = self

    def __exit__(self, *args):
        sys.stdout = self.prev


def test():
    # Testing mode
    cmd1 = ' && '.join(f'sleep 0.25 && echo "{i} "' for i in range(4))
    cmd2 = "python3 -c 'import time; [print(i, flush=True) or time.sleep(0.25) for i in range(4)] ; print(input().upper());'"
    cmd3 = "python3 -c 'import time; [print(i, flush=True) or time.sleep(0.25) for i in range(4)] ; print(input().upper()); exit(1)'"
    cmd4 = "python3 -c 'for i in range(10**6): print(str(0)*i, flush=True)'"

    class TmpWriter:

        def write(self, s):
            print(s, end='', flush=True) or time.sleep(0.6)

    tests = [
        {
            'title': 'No live printing, no error and capture stdout',
            'cmd': cmd1,
            'kwargs': dict(
                shell=True,
                timeout=None,
                capture_stdout=True,
            )
        },
        {
            'title':
                'Print 1..4 (live), no error and capture stdout',
            'cmd':
                cmd1,
            'kwargs':
                dict(
                    shell=True,
                    timeout=1.1,
                    live_stdout=sys.stdout,
                    capture_stdout=True,
                )
        },
        {
            'title':
                'Print 1..4 (live), no error and do not capture stdout',
            'cmd':
                cmd1,
            'kwargs':
                dict(
                    shell=True,
                    timeout=1.1,
                    live_stdout=sys.stdout,
                    capture_stdout=False,
                )
        },
        {
            'title':
                'Print 1..? (live), Timeout error, capture stdout',
            'cmd':
                cmd1,
            'kwargs':
                dict(
                    shell=True,
                    timeout=0.6,
                    live_stdout=sys.stdout,
                    capture_stdout=True,
                )
        },
        {
            'title':
                'Live printing, Timeout error, no capture, wait for handler',
            'cmd':
                cmd2,
            'kwargs':
                dict(
                    shell=True,
                    timeout=0.6,
                    live_stdout=TmpWriter(),
                    #wait_live_stdout=False,
                    capture_stdout=False,
                )
        },
        {
            'title': 'Live printing, Excess of Output',
            'cmd': cmd4,
            'kwargs': dict(
                shell=True,
                live_stdout=sys.stdout,
                max_stdout='1k',
            )
        },
    ]
    for i, test in enumerate(tests):
        print('-' * 10, f'TEST {i+1}', '-' * 10)
        print(test['title'])
        p = MyProcess(
            test['cmd'],
            shell=test['kwargs'].pop('shell', False),
        )
        p.run(**test['kwargs'])
        print('Elapsed:', p.elapsed)
        print('Error:', p.error)
        print('Stdout:', p.stdout)
    exit(0)  # Required for some reason



from setuptools import setup
'''
For submitting to PyPi: python3 sumbit.py
'''

setup(
    name='caph1993-pytools',
    version='0.2.20',
    description='Python toolbox of Carlos Pinzón',
    url='https://github.com/caph1993/caph1993-pytools',
    author='Carlos Pinzón',
    author_email='caph1993@gmail.com',
    license='MIT',
    packages=[
        'cp93pytools',
    ],
    install_requires=[
        'typing-extensions',
    ],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
)


import ctypes
import threading


def terminate_thread(
    thread: threading.Thread,
    exception=SystemExit,
):
    """
    https://stackoverflow.com/a/15274929/3671939
    Terminates a python thread from another thread.
    :param thread: a threading.Thread instance
    """
    if not thread.isAlive():
        return

    exc = ctypes.py_object(exception)
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
        ctypes.c_long(thread.ident or 0), exc)
    if res == 0:
        raise ValueError("nonexistent thread id")
    elif res > 1:
        # """if it returns a number greater than one, you're in trouble,
        # and you should call it again with exc=NULL to revert the effect"""
        ctypes.pythonapi.PyThreadState_SetAsyncExc(
            thread.ident,
            None,
        )
        raise SystemError("PyThreadState_SetAsyncExc failed")


if __name__ == '__main__':
    import time
    from threading import Thread

    def test1():
        for i in range(5):
            print(f'I am alive at {i}')
            time.sleep(1)
        return

    t = Thread(target=test1)
    t.start()
    time.sleep(1.5)
    terminate_thread(t, exception=KeyboardInterrupt)

    def test2():
        print('This is test2')
        t = Thread(target=test1)
        t.start()
        time.sleep(1.5)
        terminate_thread(t)
        print('End of test2')
        return

    t = Thread(target=test2)
    t.start()
    time.sleep(2.5)
    terminate_thread(t)

    print('BYE')
